{$APPTYPE CONSOLE}
program aylchk;

uses
 Windows, SysUtils, FileUtil, LazUTF8;

const
 INVALID_FILE_ATTRIBUTES = DWORD(-1); //not defined in Windows unit

type
 Available_Types =
 (Unknown, OUTFile, ZXAYFile, EPSGFile, AYFile, AYMFile, STCFile, ASCFile,
  ASC0File, STPFile, PSCFile, FLSFile, FTCFile, PT1File, PT2File, PT3File,
  SQTFile, GTRFile, FXMFile, PSMFile, VTXFile, YM2File, YM3File, YM3BFile,
  YM5File, YM6File, PSGFile, MP3File, MP2File, MP1File, OGGFile, WAVFile,
  WMAFile, MO3File, ITFile, XMFile, S3MFile, MTMFile, MODFile, UMXFile,
  CDAFile, MIDFile, RMIFile, KARFile, XMIFile);
const
 MinType = Succ(Low(Available_Types));
 MaxType = High(Available_Types);
 Version_String = 'ZX Spectrum Sound Chip Emulator Play List File v1.';
 VSLen = Length(Version_String) + 1;
 STypes:array[MinType..MaxType] of string[4] =
 ('OUT','ZXAY','EPSG','AY','AYM','STC','ASC','ASC0','STP','PSC','FLS','FTC',
  'PT1','PT2','PT3','SQT','GTR','FXM','PSM','VTX','YM2','YM3','YM3b','YM5',
  'YM6','PSG','MP3','MP2','MP1','OGG','WAV','WMA','MO3','IT','XM','S3M','MTM',
  'MOD','UMX','CDA','MID','RMI','KAR','XMI');
 SExts:array[MinType..MaxType] of string[5] =
 ('.OUT','.ZXAY','.PSG','.AY','.AYM','.STC','.ASC','.ASC','.STP','.PSC','.FLS',
  '.FTC','.PT1','.PT2','.PT3','.SQT','.GTR','.FXM','.PSM','.VTX','.YM','.YM',
  '.YM','.YM','.YM','.PSG','.MP3','.MP2','.MP1','.OGG','.WAV','.WMA','.MO3',
  '.IT','.XM','.S3M','.MTM','.MOD','.UMX','.CDA','.MID','.RMI','.KAR','.XMI');
 NumOfTokens = 21;
 MyTokens:array[0..NumOfTokens - 1] of string =
  ('ChipType','Channels','ChannelsAllocation','ChipFrequency',
   'PlayerFrequency','Offset','Length','Address','Loop','Time','Original',
   'Name','Author','Program','Computer','Date','Comment','Tracker','Type',
   'ams_andsix','FormatSpec');
 MaxTokenLen = 18;
var
 m3uf:TextFile;
 String1,String2:string;
 TokenError:boolean;
 i2:integer;
 Recurse:boolean;
 FNStr:string;
 NOfFiles:integer;

function SysErrorMessageUTF8(ErrorCode: Integer): String;
const
  MaxMsgSize = Format_Message_Max_Width_Mask;
var
  MsgBuffer: pWideChar;
begin
  GetMem(MsgBuffer, MaxMsgSize*SizeOf(WideChar));
  FillChar(MsgBuffer^, MaxMsgSize*SizeOf(WideChar), #0);
  FormatMessageW(FORMAT_MESSAGE_FROM_SYSTEM,
                 nil,
                 ErrorCode,
                 MakeLangId(LANG_NEUTRAL, SUBLANG_DEFAULT),
                 MsgBuffer,                 { This function allocs the memory }
                 MaxMsgSize,                           { Maximum message size }
                 nil);
  SysErrorMessageUTF8 := UTF8Encode(WideString(MsgBuffer));
  FreeMem(MsgBuffer, MaxMsgSize*SizeOf(WideChar));
end;

procedure ExtractToken(const S1:string;var S2:string;var Ind:integer);
var
 i:integer;
begin
i := 1;
S2 := '';
while (i <= MaxTokenLen + 1) and (i <= Length(S1)) and (S1[i] <> '=') do Inc(i);
if (i > Length(S1)) or (i > MaxTokenLen + 1) then
 begin
  TokenError := True;
  exit
 end;
S2 := Copy(S1,1,i - 1);
Ind := 0;
while (Ind < NumOfTokens) and (MyTokens[Ind] <> S2) do inc(Ind);
if Ind = NumOfTokens then
 begin
  TokenError := True;
  exit
 end;
S2 := Copy(S1,i + 1,Length(S1) - i)
end;

procedure ExtractChType(S1:string);
begin
if (S1 <> 'AY') and (S1 <> 'YM') then TokenError := True
end;

procedure ExtractChans(const S1:string);
begin
if (S1 <> 'Mono') and (S1 <> 'Stereo') then TokenError := True
end;

procedure ExtractChanMode(S1:string);
var
 i,j,k,Temp:integer;
begin
if (S1 = 'Mono') or
   (S1 = 'ABC') or (S1 = 'ACB') or (S1 = 'BAC') or
   (S1 = 'BCA') or (S1 = 'CAB') or (S1 = 'CBA') then exit;
i := 1; S1 := S1 + ',';
for j := 0 to 5 do
 begin
  k := i;
  while (i <= Length(S1)) and (S1[i] <> ',') do Inc(i);
  if i > Length(S1) then
   begin
    TokenError := True;
    exit
   end;
  Val(Copy(S1,k,i - k),k,Temp);
  if (Temp <> 0) or not (k in [0..255]) then
   begin
    TokenError := True;
    exit
   end;
  Inc(i)
 end
end;

function ExtractInteger(const S1:string):integer;
var
 Temp:integer;
begin
Val(S1,Result,Temp);
if Temp <> 0 then
 TokenError := True
end;

procedure ExtractFType(const S1:string;var FT:Available_Types);
begin
FT := MinType;
while (FT <= MaxType) and (SExts[FT] <> S1) do Inc(FT);
if FT > MaxType then
 TokenError := True
end;

procedure ExtractFType2(const S1:string;var FT:Available_Types);
begin
FT := MinType;
while (FT <= MaxType) and (STypes[FT] <> S1) do Inc(FT);
if FT > MaxType then
 TokenError := True
end;

type
 TErrors = (ErNotAYL,ErNoRecords,ErUnrKWrd,ErUnrParam,ErFileNotFound,ErUnkType,
  ErBadTS,ErBadFileSize);
const
 Errors:array[TErrors] of string =
  ('Not AYL-file','File has no records: nothing to check','Unrecognized keyword',
   'Unrecognized parameter','File not found','Unsupported file type',
   'Strange TS-link (offsets and sizes)','Strange offset+size');
var
 LineNumber:integer;

procedure ShowError(Er:TErrors);
begin
if Er = ErFileNotFound then Dec(LineNumber);
Writeln('Error in line ',LineNumber,': ',Errors[Er],'.');
Writeln('Analizing is stopped.');
Halt(Ord(Er) + 1)
end;

procedure Readln1(var Str:string);
begin
Readln(m3uf,Str);
Inc(LineNumber)
end;

procedure CheckPlaylist(const PLFName:string;FTAYL:boolean);
var
 FileType:Available_Types;
 FileName:string;
 Last,fend:boolean;
 First:integer;
 utf8:boolean;
 offs1,len1,offs2,len2:int64;
begin
Inc(NOfFiles);
SetCurrentDir(ExtractFileDir(PLFName));
Writeln('Analizing ' + PLFName + '...');
AssignFile(m3uf,PLFName);
Reset(m3uf);
LineNumber := 0;
utf8 := False;
if FTAYL then
 begin
  if eof(m3uf) then
   ShowError(ErNotAYL);
  Readln1(String1);
  if (VSLen <> Length(String1)) or
      not (String1[VSLen] in ['0'..'6']) or
      (Copy(String1,1,VSLen - 1) <> Version_String) then
   ShowError(ErNotAYL);
  utf8 := String1[VSLen] > '5';
 end;
if utf8 then
 SetTextCodePage(m3uf,CP_UTF8)
else
 SetTextCodePage(m3uf,GetACP);
TokenError := False;
if eof(m3uf) then ShowError(ErNoRecords);
ReadLn1(String1);

if FTAYL then
 begin

if String1 = '<' then
 while not eof(m3uf) do
  begin
   ReadLn1(String1);
   if String1 = '>' then
    begin
     if not eof(m3uf) then
      ReadLn1(String1)
     else
      ShowError(ErNoRecords);
     break;
    end;
   if String1 <> '' then
    begin
     ExtractToken(String1,String2,i2);
     if TokenError then
      ShowError(ErUnrKWrd);
     case i2 of
     0:   ExtractChType(String2);
     1:   ExtractChans(String2);
     2:   ExtractChanMode(String2);
     3,4: ExtractInteger(String2);
     else TokenError := True
     end;
     if TokenError then
      ShowError(ErUnrParam);
    end
  end;

 end;

Last := False;
First := -1;
while not Last do
 begin
  if First < 0 then
   begin
    offs1 := 0; offs2 := 0; len1 := 0; len2 := 0;
   end;
  inc(First);
  if First = 0 then
   if eof(m3uf) then
    begin
     String2 := '';
     Inc(LineNumber);
     Last := True;
    end
   else
    ReadLn1(String2);
  if not FTAYL or (String2 <> '<') then
   begin
    if not FileExists(String1) then
     ShowError(ErFileNotFound);
    String1 := UpperCase(ExtractFileExt(String1));
    ExtractFType(String1,FileType);
    if TokenError then
     begin
      Dec(LineNumber);
      ShowError(ErUnkType)
     end;
    String1 := String2;
    First := -1;
   end
  else if FileExists(String1) then
   begin
    FileType := Unknown;
    FileName := ExpandFileName(String1);
    if First = 0 then
     begin
      len1 := FileSize(FileName);
      if UpperCase(ExtractFileExt(FileName)) = '.FXM' then
       dec(len1,6);
     end;
    while not eof(m3uf) do
     begin
      ReadLn1(String1);
      if String1 = '>' then
       begin
        if FileType = Unknown then
         begin
          String1 := UpperCase(ExtractFileExt(FileName));
          ExtractFType(String1,FileType);
          if TokenError then ShowError(ErUnkType);
         end;
        break
       end;
      ExtractToken(String1,String2,i2);
      if TokenError then
       ShowError(ErUnrKWrd);
      case i2 of
      0:   ExtractChType(String2);
      1:   ExtractChans(String2);
      2:   ExtractChanMode(String2);
      5:   if First = 0 then
            offs1 := ExtractInteger(String2)
           else
            offs2 := ExtractInteger(String2);
      6:   if First = 0 then
            len1 := ExtractInteger(String2)
           else
            len2 := ExtractInteger(String2);
      3..4,7..10,19..21:
           ExtractInteger(String2);
      18:  ExtractFType2(String2,FileType)
      end;
      if TokenError then
       ShowError(ErUnrParam)
     end;
    fend := eof(m3uf);
    if not fend then
     Readln1(String1);
    if First = 1 then //ts
     if (FileSize(FileName) <> len1+len2+16) or
        (offs1 <> 0) or (offs2 <> len1) then
      ShowError(ErBadTS);
    if First = 0 then
     begin
      if not fend and (String1 = '<') then
       begin
        String1 := FileName;
        String2 := '<';
       end
      else
       begin
        First := -1;
        if not (
            (FileType in [AYFile,YM2File..YM6File]) or
            ((UpperCase(ExtractFileExt(FileName)) = '.FXM') and (FileSize(FileName) = offs1+len1+6)) or
            ((UpperCase(ExtractFileExt(FileName)) <> '.FXM') and (FileSize(FileName) = offs1+len1))
           ) then
         ShowError(ErBadFileSize);
       end;
     end
    else
     First := -1;
    if fend then
     break;
   end
  else
   ShowError(ErFileNotFound);
 end;
CloseFile(m3uf);
Writeln('No errors');
end;

procedure CheckPath(Dir:string);
var
 SearchRec:TSearchRec;
 i:integer;
 ext:string;
begin
Dir := IncludeTrailingBackslash(Dir);
i := FindFirst(Dir + FNStr,faAnyFile,SearchRec);
try
while i <> ERROR_NO_MORE_FILES do
 begin
  if i = 0 then
   if (SearchRec.Name <> '.') and (SearchRec.Name <> '..') then
    if SearchRec.Attr and faDirectory <> 0 then
     begin
      if Recurse then
       CheckPath(Dir + SearchRec.Name)
     end
    else if SearchRec.Size > 0 then
     begin
      ext := UpperCase(ExtractFileExt(SearchRec.Name));
      if (ext = '.AYL') or (ext = '.M3U') then
       CheckPlaylist(Dir + SearchRec.Name,ext = '.AYL');
     end;
  i := FindNext(SearchRec);
 end;
finally
FindClose(SearchRec);
end
end;

procedure ExitUsage;
 begin
  Writeln('  Usage:');
  Writeln(' aylchk [/r] [foldername]');
  Writeln('  or');
  Writeln(' aylchk [/r] [filename]');
  Writeln;
  Writeln('/r - recurse subfolders');
  Writeln('Wildcards are allowed in names.');
  Halt(1);
 end;

var
 ParStr:string;
 err:DWORD;

begin
Writeln('Playlist Files Checker');
Writeln('Supported AYL (v1.0-1.6) and M3U (only filenames)');
Writeln('(c)2002-2007,2017 S.V.Bulba E-Mail: svbulba@gmail.com');
Writeln('Support page: http://bulba.untergrund.net');
Writeln;
Recurse := False;
case ParamCount of
1:ParStr := ParamStr(1);
2:
 begin
  if LowerCase(ParamStr(1)) <> '/r' then ExitUsage;
  Recurse := True;
  ParStr := ParamStr(2);
 end
else
 ExitUsage;
end;

FileMode := 0;
err := GetFileAttributesW(PWideChar(UTF8Decode(ParStr)));
if err = INVALID_FILE_ATTRIBUTES then
 begin
  Writeln(SysErrorMessageUTF8(GetLastError));
  halt(1);
 end;
if err and FILE_ATTRIBUTE_DIRECTORY <> 0 then
 FNStr := '*.*'
else
 begin
  FNStr := ExtractFileName(ParStr);
  ParStr := ExtractFilePath(ParStr);
 end;
NOfFiles := 0;
SetCurrentDir(ExtractFileDir(ParamStr(0)));
ParStr := ExpandFileName(ParStr);
CheckPath(ParStr);
if NOfFiles = 0 then Writeln('No files are found');
end.
